import sys
import json
import textwrap
import requests
from pprint import pprint
from botocore.exceptions import ClientError

from nimbo import CONFIG
from nimbo.core.print import print


def create_security_group(group_name, dry_run=False):

    ec2 = CONFIG.get_session().client("ec2")
    response = ec2.describe_vpcs()
    vpc_id = response.get("Vpcs", [{}])[0].get("VpcId", "")

    response = ec2.create_security_group(
        GroupName=group_name,
        Description="Base VPC security group for Nimbo jobs.",
        VpcId=vpc_id,
        DryRun=dry_run,
    )

    security_group_id = response["GroupId"]
    print(
        f"Security Group {group_name} (id={security_group_id}) Created in vpc {vpc_id}."
    )


def allow_inbound_current_ip(group_name, dry_run=False):

    ec2 = CONFIG.get_session().client("ec2")

    try:
        # Get the security group id
        response = ec2.describe_security_groups(GroupNames=[group_name], DryRun=dry_run)
    except ClientError as e:
        if e.response["Error"]["Code"] == "InvalidGroup.NotFound":
            print(
                f"Security group {group_name} not found. Please use an existing security group "
                "or create a new one in the AWS console.",
                style="error",
            )
            sys.exit(1)
        elif e.response["Error"]["Code"] == "UnauthorizedOperation":
            return
        else:
            raise

    security_group_id = response["SecurityGroups"][0]["GroupId"]
    my_public_ip = requests.get("https://checkip.amazonaws.com").text.strip()

    try:
        response = ec2.authorize_security_group_ingress(
            GroupId=security_group_id,
            IpPermissions=[
                {
                    "IpProtocol": "tcp",
                    "FromPort": 22,
                    "ToPort": 22,
                    "IpRanges": [{"CidrIp": f"{my_public_ip}/16"}],
                }
            ],
        )
    except ClientError as e:
        if e.response["Error"]["Code"] == "InvalidPermission.Duplicate":
            return
        elif e.response["Error"]["Code"] == "UnauthorizedOperation":
            return
        else:
            raise


def create_instance_profile_and_role(dry_run=False):
    iam = CONFIG.get_session().client("iam")
    role_name = "NimboS3AndEC2FullAccess"
    instance_profile_name = "NimboInstanceProfile"

    policy = {
        "Version": "2012-10-17",
        "Statement": {
            "Effect": "Allow",
            "Action": "sts:AssumeRole",
            "Principal": {"Service": "ec2.amazonaws.com"},
        },
    }
    if dry_run:
        return
    iam.create_role(RoleName=role_name, AssumeRolePolicyDocument=json.dumps(policy))
    iam.attach_role_policy(
        PolicyArn="arn:aws:iam::aws:policy/AmazonS3FullAccess", RoleName=role_name
    )
    iam.attach_role_policy(
        PolicyArn="arn:aws:iam::aws:policy/AmazonEC2FullAccess", RoleName=role_name
    )

    iam.create_instance_profile(InstanceProfileName=instance_profile_name, Path="/")
    iam.add_role_to_instance_profile(
        InstanceProfileName=instance_profile_name, RoleName=role_name
    )


def create_instance_profile(role_name, dry_run=False):
    iam = CONFIG.get_session().client("iam")
    instance_profile_name = "NimboInstanceProfile"

    if dry_run:
        return

    iam.create_instance_profile(InstanceProfileName=instance_profile_name, Path="/")
    iam.add_role_to_instance_profile(
        InstanceProfileName=instance_profile_name, RoleName=role_name
    )


def list_instance_profiles(dry_run=False):
    iam = CONFIG.get_session().client("iam")

    if dry_run:
        return
    response = iam.list_instance_profiles()
    pprint(response["InstanceProfiles"])


def verify_nimbo_instance_profile(dry_run=False):
    iam = CONFIG.get_session().client("iam")

    if dry_run:
        return

    response = iam.list_instance_profiles()
    instance_profiles = response["InstanceProfiles"]
    instance_profile_names = [p["InstanceProfileName"] for p in instance_profiles]
    if "NimboInstanceProfiles" not in instance_profile_names:
        raise Exception(
            textwrap.dedent(
                """Instance profile 'NimboInstanceProfile' not found.

                An instance profile is necessary to give your instance access
                to EC2 and S3 resources. You can create an instance profile using
                'nimbo create_instance_profile <role_name>'. If you are a root user,
                you can simply run 'nimbo create_instance_profile_and_role', and
                nimbo will create the necessary role policies and instance profile
                for you. Otherwise, please ask your admin for a role that provides
                the necessary EC2 and S3 read/write access.

                For more details please go to docs.nimbo.sh/instance-profiles."
                """
            )
        )



import click

from nimbo.core import access, execute, storage, utils, setup
from nimbo.core.config import RequiredCase


@click.group(context_settings=dict(max_content_width=150))
def cli():
    pass


@cli.command()
@utils.assert_required_config(RequiredCase.NONE)
@utils.handle_errors
def generate_config():
    """Creates a base nimbo-config.yml in your current directory.

    Remember to change any fields to your own values.
    """
    utils.generate_config()


@cli.command()
@click.argument("profile")
@click.option("--full-s3-access", is_flag=True)
@utils.assert_required_config(RequiredCase.NONE)
@utils.handle_errors
def admin_setup(profile, full_s3_access):
    setup.setup(profile, full_s3_access)


@cli.command()
@click.argument("username")
@click.argument("profile")
@utils.assert_required_config(RequiredCase.NONE)
@utils.handle_errors
def add_user(username, profile):
    """Adds user USERNAME to the user group NimboUserGroup.

    You must have run 'nimbo admin-setup' before adding users.

    PROFILE is the profile name of your root/admin account.
    """
    setup.add_user(username, profile)


@cli.command()
@click.argument("job_cmd")
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.JOB)
@utils.handle_errors
def run(job_cmd, dry_run):
    """Runs the JOB_CMD command on an EC2 instance.

    JOB_CMD is any command you would run locally.\n
    E.g. \"python runner.py --epochs=10\".\n
    The command must be between quotes.
    """
    execute.run_job(job_cmd, dry_run)


@cli.command()
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.JOB)
@utils.handle_errors
def launch(dry_run):
    """
    Launches an EC2 instance according to your nimbo-config,
    without doing any further setup.
    """
    execute.run_job("_nimbo_launch", dry_run)


@cli.command()
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.JOB)
@utils.handle_errors
def launch_and_setup(dry_run):
    """
    Launches an EC2 instance with your code, data and environment,
    without running any job.
    """
    execute.run_job("_nimbo_launch_and_setup", dry_run)


@cli.command()
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.JOB)
@utils.handle_errors
def notebook(dry_run):
    """
    Launches jupyter lab on an EC2 instance with your code, data and environment.

    Make sure to run 'nimbo sync-notebooks <instance_id>' frequently to sync
    the notebook to your local folder, as the remote notebooks will be lost
    once the instance is terminated.
    """
    execute.run_job("_nimbo_notebook", dry_run)


@cli.command()
@click.argument("instance_id")
@utils.assert_required_config(RequiredCase.INSTANCE)
@utils.handle_errors
def sync_notebooks(instance_id):
    """
    Syncs the ipynb files from the instance INSTANCE_ID to your local folder.

    Make sure to run 'nimbo sync-notebooks <instance_id>' to sync the notebook
    to your local folder, as the remote notebooks will be lost once the instance
    is terminated.
    """
    execute.sync_notebooks(instance_id)


@cli.command()
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.INSTANCE, RequiredCase.STORAGE)
@utils.handle_errors
def test_access(dry_run):
    """Runs a mock job to test your config file, permissions, and credentials."""
    execute.run_access_test(dry_run)


@cli.command()
@click.argument("instance_id")
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.INSTANCE)
@utils.handle_errors
def ssh(instance_id, dry_run):
    """SSH into an instance by INSTANCE_ID."""
    utils.ssh(instance_id, dry_run)


@cli.command()
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def list_gpu_prices(dry_run):
    """Lists the prices, types, and specs of GPU instances."""
    utils.list_gpu_prices(dry_run)


@cli.command()
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def list_spot_gpu_prices(dry_run):
    """Lists the prices, types, and specs of GPU spot instances."""
    utils.list_spot_gpu_prices(dry_run)


@cli.command()
@click.argument("qty", type=int, required=True)
@click.argument("timescale", type=click.Choice(["days", "months"]), required=True)
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def spending(qty, timescale, dry_run):
    """Shows daily/monthly spending summary. Costs without credits or refunds applied.

    QTY is the number of days/months you want to see, starting from the current date.\n
    For example:\n
        'nimbo spending 10 days' will show daily spending of the last 10 days\n
        'nimbo spending 3 months' will show the monthly spending of the last 3 months
    """
    utils.show_spending(qty, timescale, dry_run)


@cli.command()
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def list_active(dry_run):
    """Lists all your active instances."""
    utils.show_active_instances(dry_run)


@cli.command()
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def list_stopped(dry_run):
    """Lists all your stopped instances."""
    utils.show_stopped_instances(dry_run)


@cli.command()
@click.argument("instance_id")
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def check_instance_status(instance_id, dry_run):
    """Checks the status of an instance by INSTANCE_ID."""
    utils.check_instance_status(instance_id, dry_run)


@cli.command()
@click.argument("instance_id")
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def stop_instance(instance_id, dry_run):
    """Stops an instance by INSTANCE_ID."""
    utils.stop_instance(instance_id, dry_run)


@cli.command()
@click.argument("instance_id")
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def delete_instance(instance_id, dry_run):
    """Terminates an instance by INSTANCE_ID."""
    utils.delete_instance(instance_id, dry_run)


@cli.command()
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def delete_all_instances(dry_run):
    """Terminates all your instances."""
    click.confirm(
        "This will delete all your running instances.\n" "Do you want to continue?",
        abort=True,
    )
    utils.delete_all_instances(dry_run)


@cli.command()
@click.argument("bucket_name")
@click.option("--dry-run", is_flag=True)
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def create_bucket(bucket_name, dry_run):
    """
    Create a bucket BUCKET_NAME in S3.

    BUCKET_NAME is the name of the bucket to create, s3://BUCKET_NAME
    """
    storage.create_bucket(bucket_name, dry_run)


@cli.command()
@click.argument(
    "folder", type=click.Choice(["datasets", "results", "logs"]), required=True
)
@click.option(
    "--delete",
    is_flag=True,
    help="""
      Deletes any files that exist in the local folder
      but don't exist in the remote folder.
    """,
)
@utils.assert_required_config(RequiredCase.STORAGE)
@utils.handle_errors
def push(folder, delete):
    """Push your local datasets/results folder onto S3."""

    if delete:
        click.confirm(
            "This will delete any files that exist in the remote "
            "folder but do not exist in the local folder.\n"
            "Do you want to continue?",
            abort=True,
        )
    storage.push(folder, delete)


@cli.command()
@click.argument(
    "folder", type=click.Choice(["datasets", "results", "logs"]), required=True
)
@click.option(
    "--delete",
    is_flag=True,
    help="""
      Deletes any files that exist in the local
      folder but don't exist in the remote folder.
    """,
)
@utils.assert_required_config(RequiredCase.STORAGE)
@utils.handle_errors
def pull(folder, delete):
    """Pull the S3 datasets/results folder into your local computer."""

    if delete:
        click.confirm(
            "This will delete any files that exist in the local "
            "folder but do not exist in the remote folder.\n"
            "Do you want to continue?",
            abort=True,
        )
    storage.pull(folder, delete)


@cli.command()
@click.argument("path")
@utils.assert_required_config(RequiredCase.MINIMAL)
@utils.handle_errors
def ls(path):
    """List the s3 objects in PATH.

    PATH is an s3 path of the form s3://bucket-name/my/files/path.
    """
    storage.ls(path)



import functools
import json
import os
import subprocess
import sys
from pprint import pprint

from datetime import datetime, timedelta, date
from dateutil.relativedelta import relativedelta

import botocore
import botocore.errorfactory
import requests
from botocore.exceptions import ClientError

from nimbo import CONFIG
from nimbo.core.config import RequiredCase
from nimbo.core.environment import is_test_environment
from nimbo.core.statics import FULL_REGION_NAMES, INSTANCE_GPU_MAP, NIMBO_DEFAULT_CONFIG
from nimbo.core.print import print, print_header


def ec2_instance_types():
    """Yield all available EC2 instance types in region CONFIG.region_name"""
    describe_args = {}
    client = CONFIG.get_session().client("ec2")
    while True:
        describe_result = client.describe_instance_types(**describe_args)
        yield from [i["InstanceType"] for i in describe_result["InstanceTypes"]]
        if "NextToken" not in describe_result:
            break
        describe_args["NextToken"] = describe_result["NextToken"]


def format_price_string(instance_type, price, gpus, cpus, mem):
    string = "\t{0: <16} {1: <15} {2: <10} {3: <5} {4:<7}".format(
        instance_type, price, gpus, cpus, mem
    )
    return string


def list_gpu_prices(dry_run=False):
    if dry_run:
        return

    instance_types = list(sorted(ec2_instance_types()))
    instance_types = [
        inst
        for inst in instance_types
        if inst[:2] in ["p2", "p3", "p4"] or inst[:3] in ["g4d"]
    ]
    full_region_name = FULL_REGION_NAMES[CONFIG.region_name]

    pricing = CONFIG.get_session().client("pricing", region_name="us-east-1")

    string = format_price_string(
        "InstanceType", "Price ($/hour)", "GPUs", "CPUs", "Mem (Gb)"
    )
    print()
    print(string, style="bold")

    for instance_type in instance_types:
        response = pricing.get_products(
            ServiceCode="AmazonEC2",
            MaxResults=100,
            FormatVersion="aws_v1",
            Filters=[
                {"Type": "TERM_MATCH", "Field": "instanceType", "Value": instance_type},
                {"Type": "TERM_MATCH", "Field": "location", "Value": full_region_name},
                {"Type": "TERM_MATCH", "Field": "operatingSystem", "Value": "Linux"},
                {"Type": "TERM_MATCH", "Field": "capacitystatus", "Value": "Used"},
                {"Type": "TERM_MATCH", "Field": "preInstalledSw", "Value": "NA"},
                {"Type": "TERM_MATCH", "Field": "tenancy", "Value": "shared"},
            ],
        )

        inst = json.loads(response["PriceList"][0])
        inst = inst["terms"]["OnDemand"]
        inst = list(inst.values())[0]
        inst = list(inst["priceDimensions"].values())[0]
        inst = inst["pricePerUnit"]
        currency = list(inst.keys())[0]
        price = float(inst[currency])

        num_gpus, gpu_type, mem, cpus = INSTANCE_GPU_MAP[instance_type]
        string = format_price_string(
            instance_type, round(price, 2), f"{num_gpus} x {gpu_type}", cpus, mem
        )
        print(string)
    print()


def list_spot_gpu_prices(dry_run=False):
    if dry_run:
        return

    instance_types = list(sorted(ec2_instance_types()))
    instance_types = [
        inst
        for inst in instance_types
        if inst[:2] in ["p2", "p3", "p4"] or inst[:3] in ["g4d"]
    ]

    ec2 = CONFIG.get_session().client("ec2")

    string = format_price_string(
        "InstanceType", "Price ($/hour)", "GPUs", "CPUs", "Mem (Gb)"
    )
    print()
    print(string, style="bold")

    for instance_type in instance_types:
        response = ec2.describe_spot_price_history(
            InstanceTypes=[instance_type],
            Filters=[{"Name": "product-description", "Values": ["Linux/UNIX"]}],
        )

        price = float(response["SpotPriceHistory"][0]["SpotPrice"])

        num_gpus, gpu_type, mem, cpus = INSTANCE_GPU_MAP[instance_type]
        string = format_price_string(
            instance_type, round(price, 2), f"{num_gpus} x {gpu_type}", cpus, mem
        )
        print(string)
    print()


def show_spending(qty, timescale, dry_run=False):

    today = date.today()
    timeformat = "%Y-%m-%d"

    if timescale == "months":
        start_date = (today - relativedelta(months=qty)).replace(day=1)
        start = start_date.strftime(timeformat)
        end = today.strftime(timeformat)
        granularity = "monthly"
    elif timescale == "days":
        start_date = today - relativedelta(days=qty)
        start = start_date.strftime(timeformat)
        end = today.strftime(timeformat)
        granularity = "daily"
    else:
        raise ValueError("Timescale must be 'daily' or 'monthly'.")

    services = [
        "Amazon Elastic Compute Cloud - Compute",
        "EC2 - Other",
        "Amazon Simple Storage Service",
    ]

    client = CONFIG.get_session().client("ce")
    results = client.get_cost_and_usage(
        TimePeriod={"End": end, "Start": start},
        Granularity=granularity.upper(),
        Filter={
            "And": [
                {
                    "Not": {
                        "Dimensions": {
                            "Key": "RECORD_TYPE",
                            "Values": ["Credit", "Refund"],
                        }
                    }
                },
                {
                    "Dimensions": {
                        "Key": "SERVICE",
                        "Values": services,
                    }
                },
            ]
        },
        Metrics=["UnblendedCost"],
        GroupBy=[{"Type": "DIMENSION", "Key": "SERVICE"}],
    )["ResultsByTime"]

    table = []
    print()
    for interval in results:
        period_start_date = datetime.strptime(
            interval["TimePeriod"]["Start"], "%Y-%m-%d"
        )
        if granularity == "monthly":
            period_string = period_start_date.strftime("%b %Y")
        else:
            period_string = period_start_date.strftime("%d %b")

        groups = interval["Groups"]
        if groups == []:
            table.append([period_string, 0, 0])
        else:
            ec2_cost = 0
            s3_cost = 0
            for group in groups:
                group_cost = float(group["Metrics"]["UnblendedCost"]["Amount"])
                if group["Keys"] == ["Amazon Elastic Compute Cloud - Compute"]:
                    ec2_cost += group_cost
                elif group["Keys"] == ["EC2 - Other"]:
                    ec2_cost += group_cost
                elif group["Keys"] == ["Amazon Simple Storage Service"]:
                    s3_cost += group_cost

            table.append([period_string, ec2_cost, s3_cost])

    def row_string(x):
        string = f"\t{x[0]:>10}"
        for xi in x[1:]:
            if type(xi) == float:
                string += f" {xi:>10.2f}"
            else:
                string += f" {xi:>10}"
        return string

    print(f"\tSpending for region {CONFIG.region_name}:")
    print()
    print(row_string(["", "EC2", "S3"]))
    for row in table:
        # round before passing to row_string
        print(row_string(row))

    ec2_total = sum([row[1] for row in table])
    s3_total = sum([row[2] for row in table])
    print("\t" + "-" * 32)
    print(row_string(["Total", ec2_total, s3_total]))
    print()


def show_active_instances(dry_run=False):
    ec2 = CONFIG.get_session().client("ec2")
    try:
        response = ec2.describe_instances(
            Filters=[{"Name": "instance-state-name", "Values": ["running", "pending"]}]
            + make_instance_filters(),
            DryRun=dry_run,
        )
        for reservation in response["Reservations"]:
            for inst in reservation["Instances"]:
                print(
                    f"Id: [bright_green]{inst['InstanceId']}[/bright_green]\n"
                    f"Status: {inst['State']['Name']}\n"
                    f"Launch Time: {inst['LaunchTime']}\n"
                    f"InstanceType: {inst['InstanceType']}\n"
                    f"IP Address: {inst['PublicIpAddress']}\n"
                )

    except ClientError as e:
        if "DryRunOperation" not in str(e):
            raise


def show_stopped_instances(dry_run=False):
    ec2 = CONFIG.get_session().client("ec2")
    try:
        response = ec2.describe_instances(
            Filters=[{"Name": "instance-state-name", "Values": ["stopped", "stopping"]}]
            + make_instance_filters(),
            DryRun=dry_run,
        )
        for reservation in response["Reservations"]:
            for inst in reservation["Instances"]:
                print(
                    f"ID: {inst['InstanceId']}\n"
                    f"Launch Time: {inst['LaunchTime']}\n"
                    f"InstanceType: {inst['InstanceType']}\n"
                )
    except ClientError as e:
        if "DryRunOperation" not in str(e):
            raise


def check_instance_status(instance_id, dry_run=False):
    ec2 = CONFIG.get_session().client("ec2")
    try:
        response = ec2.describe_instances(
            InstanceIds=[instance_id], Filters=make_instance_filters(), DryRun=dry_run
        )
        status = response["Reservations"][0]["Instances"][0]["State"]["Name"]
        return status
    except ClientError as e:
        if "DryRunOperation" not in str(e):
            raise


def stop_instance(instance_id, dry_run=False):
    ec2 = CONFIG.get_session().client("ec2")
    try:
        response = ec2.stop_instances(InstanceIds=[instance_id], DryRun=dry_run)
        pprint(response)
    except ClientError as e:
        if "DryRunOperation" not in str(e):
            raise


def delete_instance(instance_id, dry_run=False):
    ec2 = CONFIG.get_session().client("ec2")
    try:
        response = ec2.terminate_instances(InstanceIds=[instance_id], DryRun=dry_run)
        status = response["TerminatingInstances"][0]["CurrentState"]["Name"]
        print_header(f"Instance [green]{instance_id}[/green]: {status}")
    except ClientError as e:
        if "DryRunOperation" not in str(e):
            raise


def delete_all_instances(dry_run=False):
    ec2 = CONFIG.get_session().client("ec2")
    try:
        response = ec2.describe_instances(
            Filters=[{"Name": "instance-state-name", "Values": ["running"]}]
            + make_instance_filters(),
            DryRun=dry_run,
        )
        for reservation in response["Reservations"]:
            for inst in reservation["Instances"]:
                instance_id = inst["InstanceId"]
                delete_response = ec2.terminate_instances(
                    InstanceIds=[instance_id],
                )
                status = delete_response["TerminatingInstances"][0]["CurrentState"][
                    "Name"
                ]
                print_header(f"Instance [green]{instance_id}[/green]: {status}")
    except ClientError as e:
        if "DryRunOperation" not in str(e):
            raise


def check_instance_host(instance_id, dry_run=False):
    ec2 = CONFIG.get_session().client("ec2")
    try:
        response = ec2.describe_instances(
            InstanceIds=[instance_id],
            Filters=make_instance_filters(),
            DryRun=dry_run,
        )
        host = response["Reservations"][0]["Instances"][0]["PublicIpAddress"]
    except ClientError as e:
        if "DryRunOperation" not in str(e):
            raise
        host = "random_host"
    return host


def list_active_buckets():
    s3 = CONFIG.get_session().client("s3")
    response = s3.list_buckets()
    pprint(response)


def ssh(instance_id, dry_run=False):
    host = check_instance_host(instance_id, dry_run)

    if dry_run:
        return

    subprocess.Popen(
        f"ssh -i {CONFIG.instance_key} "
        f"-o 'StrictHostKeyChecking no' -o ServerAliveInterval=20 "
        f"ubuntu@{host}",
        shell=True,
    ).communicate()


def make_instance_tags():
    return [
        {"Key": "CreatedBy", "Value": "nimbo"},
        {"Key": "Owner", "Value": CONFIG.user_id},
    ]


def make_instance_filters():
    tags = make_instance_tags()
    filters = []
    for tag in tags:
        tag_filter = {"Name": "tag:" + tag["Key"], "Values": [tag["Value"]]}
        filters.append(tag_filter)
    return filters


def get_image_id():
    if CONFIG.image[:4] == "ami-":
        image_id = CONFIG.image
    else:
        response = requests.get(
            "https://nimboami-default-rtdb.firebaseio.com/images.json"
        )
        catalog = response.json()
        region = CONFIG.region_name
        if region in catalog:
            region_catalog = catalog[region]
            image_name = CONFIG.image
            if image_name in region_catalog:
                image_id = region_catalog[image_name]
            else:
                raise ValueError(
                    f"The image {image_name} was not found in the"
                    " image catalog managed by Nimbo.\n"
                    "Check https://docs.nimbo.sh/managed-images"
                    " for the list of managed images."
                )
        else:
            raise ValueError(
                f"Currently, Nimbo does not support managed images in {region}."
                " Please use another region."
            )

    return image_id


def assert_required_config(*cases: RequiredCase):
    """
    Decorator for ensuring that required config is present
    """

    def decorator(func):
        @functools.wraps(func)
        def decorated(*args, **kwargs):
            try:
                CONFIG.assert_required_config_exists(*cases)
                return func(*args, **kwargs)
            except AssertionError as e:
                print(e, style="error")
                sys.exit(1)
            except FileNotFoundError as e:
                # Happens when nimbo config file is not found
                print(e, style="error")
                sys.exit(1)

        return decorated

    return decorator


def handle_errors(func):
    """
    Decorator for catching boto3 ClientErrors, ValueError or KeyboardInterrupts.
    In case of error print the error message and stop Nimbo.
    """

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        if is_test_environment():
            return func(*args, **kwargs)
        else:
            try:
                return func(*args, **kwargs)
            except botocore.errorfactory.ClientError as e:
                print(e, style="error")
                sys.exit(1)
            except ValueError as e:
                print(e, style="error")
                sys.exit(1)
            except KeyboardInterrupt:
                print("Aborting...")
                sys.exit(1)

    return decorated


def generate_config(quiet=False) -> None:
    """ Create an example Nimbo config in the project root """

    if os.path.isfile(CONFIG.nimbo_config_file):
        print(
            f"{CONFIG.nimbo_config_file} already exists, do you want to overwrite it?"
        )

        if not _get_user_confirmation():
            print("Leaving Nimbo config intact")
            return

    with open(CONFIG.nimbo_config_file, "w") as f:
        f.write(NIMBO_DEFAULT_CONFIG)

    if not quiet:
        print(f"Example config written to {CONFIG.nimbo_config_file}")


def _get_user_confirmation() -> bool:
    try:
        confirmation = input("Type Y for yes or N for no - ")
        return confirmation.lower() == "y" or confirmation.lower() == "yes"
    except BaseException as e:
        print(e, style="error")
        print("Aborting...")



import sys
import json
import boto3
from botocore.exceptions import ClientError

from nimbo import CONFIG
from nimbo.core.print import print, print_header
from nimbo.core.statics import EC2_POLICY_JSON, ASSUME_ROLE_POLICY

NIMBO_USER_GROUP = "NimboUserGroup"
EC2_POLICY_NAME = "NimboEC2Policy"
CRED_POLICY_NAME = "NimboCredentialsPolicy"
PASS_ROLE_POLICY_NAME = "NimboPassRolePolicy"
S3_ACCESS_ROLE_NAME = "NimboFullS3AccessRole"


def create_group(client, group_name):
    try:
        client.create_group(GroupName=group_name)
    except ClientError as e:
        if e.response["Error"]["Code"] == "EntityAlreadyExists":
            print(f"User group {group_name} already exists. Skipping.", style="warning")
        else:
            raise


def create_policy(client, policy_name, policy_json):
    try:
        ec2_policy = client.create_policy(
            PolicyName=policy_name, PolicyDocument=json.dumps(policy_json)
        )
    except ClientError as e:
        if e.response["Error"]["Code"] == "EntityAlreadyExists":
            print(f"Policy {policy_name} already exists. Skipping.", style="warning")
        else:
            raise


def create_role_and_instance_profile(client, role_name):
    try:
        client.create_role(
            RoleName=role_name, AssumeRolePolicyDocument=json.dumps(ASSUME_ROLE_POLICY)
        )
    except ClientError as e:
        if e.response["Error"]["Code"] == "EntityAlreadyExists":
            print(f"Role {role_name} already exists. Skipping.", style="warning")
        else:
            raise

    try:
        client.create_instance_profile(InstanceProfileName=role_name, Path="/")
    except ClientError as e:
        if e.response["Error"]["Code"] == "EntityAlreadyExists":
            print(
                f"Instance profile for role {role_name} already exists. Skipping.",
                style="warning",
            )
        else:
            raise

    try:
        client.add_role_to_instance_profile(
            InstanceProfileName=role_name, RoleName=role_name
        )
    except ClientError as e:
        if e.response["Error"]["Code"] == "LimitExceeded":
            print(
                f"Instance profile {role_name} already has a role. Skipping.",
                style="warning",
            )
        else:
            raise


def setup(profile, full_s3_access=False):
    session = boto3.Session(profile_name=profile)
    account = session.client("sts").get_caller_identity()["Account"]

    iam = session.client("iam")

    print_header(f"Creating user group {NIMBO_USER_GROUP}...")
    create_group(iam, NIMBO_USER_GROUP)

    print_header(f"Creating policy {EC2_POLICY_NAME}...")
    create_policy(iam, EC2_POLICY_NAME, EC2_POLICY_JSON)

    print_header(
        f"Attaching policy {EC2_POLICY_NAME} to user group {NIMBO_USER_GROUP}..."
    )
    iam.attach_group_policy(
        GroupName=NIMBO_USER_GROUP,
        PolicyArn=f"arn:aws:iam::{account}:policy/{EC2_POLICY_NAME}",
    )

    if full_s3_access:
        print_header(f"Creating role {S3_ACCESS_ROLE_NAME}...")
        create_role_and_instance_profile(iam, S3_ACCESS_ROLE_NAME)

        print_header(
            f"Attaching AmazonS3FullAccess policy to role {S3_ACCESS_ROLE_NAME}..."
        )
        iam.attach_role_policy(
            PolicyArn="arn:aws:iam::aws:policy/AmazonS3FullAccess",
            RoleName=S3_ACCESS_ROLE_NAME,
        )

        print_header(f"Creating policy {PASS_ROLE_POLICY_NAME}...")
        pass_role_policy_json = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "NimboPassRolePolicy",
                    "Effect": "Allow",
                    "Action": "iam:PassRole",
                    "Resource": f"arn:aws:iam::*:role/{S3_ACCESS_ROLE_NAME}",
                }
            ],
        }
        create_policy(iam, PASS_ROLE_POLICY_NAME, pass_role_policy_json)

        print_header(
            f"Attaching policy {PASS_ROLE_POLICY_NAME} to user group {NIMBO_USER_GROUP}..."
        )
        iam.attach_group_policy(
            GroupName=NIMBO_USER_GROUP,
            PolicyArn=f"arn:aws:iam::{account}:policy/{PASS_ROLE_POLICY_NAME}",
        )

        print_header(
            f"Attaching policy AmazonS3FullAccess to user group {NIMBO_USER_GROUP}..."
        )
        iam.attach_group_policy(
            GroupName=NIMBO_USER_GROUP,
            PolicyArn="arn:aws:iam::aws:policy/AmazonS3FullAccess",
        )

    else:
        print(
            "\nSince you chose not to give full S3 access to the Nimbo user group and instance role,\n"
            "we recommend that you create a role with the necessary S3 permissions in the AWS console.\n"
            "Once you do this, give the role name to the people using Nimbo so that they can set\n"
            "the 'role' field in the nimbo-config.yml to this value.",
            style="warning",
        )

    print()
    print_header("Done.")
    print_header(
        "To add users to the NimboUserGroup, simply run 'nimbo add-user USERNAME YOUR_AWS_PROFILE'.\n"
        "For more info use 'nimbo add-user --help'"
    )


def add_user(username, profile):
    session = boto3.Session(profile_name=profile)
    iam = session.client("iam")

    response = iam.add_user_to_group(GroupName=NIMBO_USER_GROUP, UserName=username)
    print(f"User {username} added to {NIMBO_USER_GROUP}.")



import os
import socket
import subprocess
import sys
import time
from os.path import join
from subprocess import PIPE
from pathlib import Path

from nimbo import CONFIG
from nimbo.core import telemetry, utils, access
from nimbo.core.storage import s3_cp_command, s3_sync_command
from nimbo.core.statics import NIMBO_ROOT, NIMBO_VARS
from nimbo.core.print import print, print_header


def launch_instance(client):

    access.allow_inbound_current_ip(CONFIG.security_group)
    image = utils.get_image_id()
    print_header(f"Launching instance with image {image}... ")

    ebs_config = {
        "VolumeSize": CONFIG.disk_size,
        "VolumeType": CONFIG.disk_type,
    }
    if CONFIG.disk_iops:
        ebs_config["Iops"] = CONFIG.disk_iops

    instance_config = {
        "BlockDeviceMappings": [{"DeviceName": "/dev/sda1", "Ebs": ebs_config}],
        "ImageId": image,
        "InstanceType": CONFIG.instance_type,
        "KeyName": Path(CONFIG.instance_key).stem,
        "Placement": {"Tenancy": "default"},
        "SecurityGroups": [CONFIG.security_group],
        "IamInstanceProfile": {"Name": CONFIG.role},
    }

    if CONFIG.spot:
        extra_kwargs = {}
        if CONFIG.spot_duration:
            extra_kwargs = {"BlockDurationMinutes": CONFIG.spot_duration}

        instance = client.request_spot_instances(
            LaunchSpecification=instance_config,
            TagSpecifications=[
                {
                    "ResourceType": "spot-instances-request",
                    "Tags": utils.make_instance_tags(),
                }
            ],
            **extra_kwargs,
        )
        instance_request = instance["SpotInstanceRequests"][0]
        request_id = instance_request["SpotInstanceRequestId"]

        try:
            print_header("Spot instance request submitted.")
            print_header("Waiting for the spot instance request to be fulfilled... ")

            status = ""
            while status != "fulfilled":
                time.sleep(2)
                response = client.describe_spot_instance_requests(
                    SpotInstanceRequestIds=[request_id],
                    Filters=utils.make_instance_filters(),
                )
                instance_request = response["SpotInstanceRequests"][0]
                status = instance_request["Status"]["Code"]
                if status not in [
                    "fulfilled",
                    "pending-evaluation",
                    "pending-fulfillment",
                ]:
                    raise Exception(response["SpotInstanceRequests"][0]["Status"])
        except KeyboardInterrupt:
            client.cancel_spot_instance_requests(SpotInstanceRequestIds=[request_id])
            print_header("Cancelled spot instance request.")
            sys.exit(1)

        print_header("Done.")
        client.create_tags(
            Resources=[instance_request["InstanceId"]], Tags=utils.make_instance_tags()
        )
        instance = instance_request
    else:
        instance_config["MinCount"] = 1
        instance_config["MaxCount"] = 1
        instance_config["InstanceInitiatedShutdownBehavior"] = "terminate"
        instance_config["TagSpecifications"] = [
            {"ResourceType": "instance", "Tags": utils.make_instance_tags()}
        ]
        instance = client.run_instances(**instance_config)
        instance = instance["Instances"][0]

    return instance


def write_nimbo_vars():
    var_list = [
        f"S3_DATASETS_PATH={CONFIG.s3_datasets_path}",
        f"S3_RESULTS_PATH={CONFIG.s3_results_path}",
        f"LOCAL_DATASETS_PATH={CONFIG.local_datasets_path}",
        f"LOCAL_RESULTS_PATH={CONFIG.local_results_path}",
    ]
    if CONFIG.encryption:
        var_list.append(f"ENCRYPTION={CONFIG.encryption}")
    with open(NIMBO_VARS, "w") as f:
        f.write("\n".join(var_list))


def wait_for_instance_running(instance_id):
    status = ""
    while status != "running":
        time.sleep(1)
        status = utils.check_instance_status(instance_id)


def block_until_ssh_ready(host: str) -> None:
    print_header(
        f"Waiting for instance to be ready for ssh at {host}. "
        "This can take up to 2 minutes... "
    )

    start = time.monotonic()

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(1)

    reconnect_count = 0
    while reconnect_count < CONFIG.ssh_timeout:
        error_num = sock.connect_ex((host, 22))

        if error_num == 0:
            break

        time.sleep(1)
        reconnect_count += 1
    else:
        raise RuntimeError(
            "Something went wrong while connecting to the instance.\n"
            "Please verify your security groups, instance key and "
            "instance profile, and try again.\n"
            "More info at docs.nimbo.sh/common-issues#cant-ssh.\n"
        )

    print_header(f"Ready. (%0.3f s)" % (time.monotonic() - start))


def sync_code(host):
    if ".git" not in os.listdir():
        print(
            "No git repo found. Syncing all the python and bash files as a fallback.",
            style="warning",
        )
        print("Please consider using git to track the files to sync.", style="warning")
        subprocess.Popen(
            f"rsync -avm -e 'ssh -i {CONFIG.instance_key}' "
            f"--include '*/' --include '*.py' --include '*.ipynb' --include '*.sh' --exclude '*' "
            f". ubuntu@{host}:/home/ubuntu/project",
            shell=True,
        ).communicate()
    else:
        output, error = subprocess.Popen(
            "git ls-tree -r HEAD --name-only", stdout=PIPE, shell=True
        ).communicate()
        git_tracked_files = output.decode("utf-8").strip().splitlines()
        include_files = [f"--include '{file_name}'" for file_name in git_tracked_files]
        include_string = " ".join(include_files)
        subprocess.Popen(
            f"rsync -amr -e 'ssh -i {CONFIG.instance_key}' "
            f"--include '*/' {include_string} --exclude '*' "
            f". ubuntu@{host}:/home/ubuntu/project",
            shell=True,
        ).communicate()


def sync_notebooks(instance_id):
    host = utils.check_instance_host(instance_id)

    subprocess.Popen(
        f"rsync -avm -e 'ssh -i {CONFIG.instance_key}' "
        f"--include '*/' --include '*.ipynb' --exclude '*' "
        f"ubuntu@{host}:/home/ubuntu/project/ .",
        shell=True
    ).communicate()


def run_remote_script(ssh_cmd, scp_cmd, host, instance_id, job_cmd, script):
    remote_script = join(NIMBO_ROOT, "scripts", script)
    subprocess.check_output(
        f"{scp_cmd} {remote_script} ubuntu@{host}:/home/ubuntu/", shell=True
    )

    nimbo_log = "/home/ubuntu/nimbo-log.txt"
    bash_cmd = f"bash {script}"
    if CONFIG.run_in_background:
        full_command = (
            f"nohup {bash_cmd} {instance_id} {job_cmd} </dev/null >{nimbo_log} 2>&1 &"
        )
    else:
        full_command = f"{bash_cmd} {instance_id} {job_cmd}"

    subprocess.Popen(
        f'{ssh_cmd} ubuntu@{host} "{full_command}"', shell=True
    ).communicate()


def run_job(job_cmd, dry_run=False):
    if dry_run:
        return {"message": job_cmd + "_dry_run"}

    # access.verify_nimbo_instance_profile(session)

    # Launch instance with new volume for anaconda
    ec2 = CONFIG.get_session().client("ec2")
    telemetry.record_event("run")

    start_t = time.monotonic()

    instance = launch_instance(ec2)
    instance_id = instance["InstanceId"]

    try:
        # Wait for the instance to be running
        wait_for_instance_running(instance_id)
        end_t = time.monotonic()
        print_header(f"Instance running. ({round((end_t-start_t), 2)} s)")
        print_header(f"InstanceId: [green]{instance_id}[/green]")
        print()

        time.sleep(5)
        host = utils.check_instance_host(instance_id)

        block_until_ssh_ready(host)

        if job_cmd == "_nimbo_launch":
            print_header(
                f"Run [cyan]nimbo ssh {instance_id}[/cyan] to log onto the instance"
            )
            return {"message": job_cmd + "_success", "instance_id": instance_id}

        ssh = (
            f"ssh -i {CONFIG.instance_key} -o 'StrictHostKeyChecking no'"
            " -o ServerAliveInterval=5 "
        )
        scp = f"scp -i {CONFIG.instance_key} -o 'StrictHostKeyChecking no'"

        local_env = "/tmp/local_env.yml"
        user_conda_yml = CONFIG.conda_env
        # TODO: Replace this with shutil
        subprocess.check_output(f"cp {user_conda_yml} {local_env}", shell=True)

        # Send conda env yaml and setup scripts to instance
        print()
        print_header(f"Syncing conda, config, and setup files...")
        write_nimbo_vars()

        # Create project folder and send env and config files there
        subprocess.check_output(f"{ssh} ubuntu@{host} mkdir project", shell=True)
        subprocess.check_output(
            f"{scp} {local_env} {CONFIG.nimbo_config_file} {NIMBO_VARS}"
            f" ubuntu@{host}:/home/ubuntu/project/",
            shell=True,
        )

        # Sync code with instance
        print()
        print_header(f"Syncing code...")
        sync_code(host)

        print_header(f"Running setup code on the instance from here on.")
        # Run remote_setup script on instance
        run_remote_script(ssh, scp, host, instance_id, job_cmd, "remote_setup.sh")

        if job_cmd == "_nimbo_notebook":
            subprocess.Popen(
                f"{ssh} -o 'ExitOnForwardFailure yes' "
                f"ubuntu@{host} -NfL 57467:localhost:57467 >/dev/null 2>&1", 
                shell=True).communicate()
            print_header(
                "Make sure to run 'nimbo sync-notebooks <instance_id>' frequently to sync "
                "the notebook to your local folder, as the remote notebooks will be lost "
                "once the instance is terminated."
            )
        return {"message": job_cmd + "_success", "instance_id": instance_id}

    except BaseException as e:
        if type(e) != KeyboardInterrupt and type(e) != subprocess.CalledProcessError:
            print(e, style="error")

        if not CONFIG.persist:
            print_header(f"Deleting instance {instance_id} (from local)... ")
            utils.delete_instance(instance_id)

        return {"message": job_cmd + "_error", "instance_id": instance_id}


def run_access_test(dry_run=False):
    if dry_run:
        return

    CONFIG.instance_type = "t3.medium"
    CONFIG.run_in_background = False
    CONFIG.persist = False

    try:
        # Send test file to s3 results path and delete it
        profile = CONFIG.aws_profile
        region = CONFIG.region_name
        results_path = CONFIG.s3_results_path

        subprocess.check_output(
            "echo 'Hello World' > nimbo-access-test.txt", shell=True
        )

        command = s3_cp_command("nimbo-access-test.txt", results_path)
        subprocess.check_output(command, shell=True)

        command = f"aws s3 ls {results_path} --profile {profile} --region {region}"
        subprocess.check_output(command, shell=True)
        command = (
            f"aws s3 rm {results_path}/nimbo-access-test.txt "
            f"--profile {profile} --region {region}"
        )
        subprocess.check_output(command, shell=True)

        print(
            "You have the necessary S3 read/write permissions from your computer \u2713"
        )

    except subprocess.CalledProcessError as e:
        print(e, style="error")
        sys.exit(1)

    # access.verify_nimbo_instance_profile(session)
    # print("Instance profile 'NimboInstanceProfile' found \u2713")

    # Launch instance with new volume for anaconda
    print("Launching test instance... ")
    ec2 = CONFIG.get_session().client("ec2")

    instance = launch_instance(ec2)
    instance_id = instance["InstanceId"]

    try:
        # Wait for the instance to be running
        wait_for_instance_running(instance_id)
        print(f"Instance running. Instance creation allowed \u2713")
        print(f"InstanceId: {instance_id}")
        print()

        print("Trying to delete this instance...")
        utils.delete_instance(instance_id)

        print("Instance deletion allowed \u2713")
        print("\nLaunching another instance...")
        instance = launch_instance(ec2)
        instance_id = instance["InstanceId"]
        print(f"Instance running. InstanceId: {instance_id}")

        time.sleep(5)
        host = utils.check_instance_host(instance_id)
        ssh = (
            f"ssh -i {CONFIG.instance_key} -o 'StrictHostKeyChecking no' "
            "-o ServerAliveInterval=20"
        )
        scp = f"scp -i {CONFIG.instance_key} -o 'StrictHostKeyChecking no'"

        block_until_ssh_ready(host)

        print("Instance key allows ssh access to remote instance \u2713")
        print("Security group allows ssh access to remote instance \u2713")

        write_nimbo_vars()

        subprocess.check_output(
            f"{scp} {CONFIG.nimbo_config_file} {NIMBO_VARS} ubuntu@{host}:/home/ubuntu/",
            shell=True,
        )
        run_remote_script(ssh, scp, host, instance_id, "", "remote_s3_test.sh")

    except BaseException as e:
        if type(e) != KeyboardInterrupt and type(e) != subprocess.CalledProcessError:
            print(e, style="error")

        if not CONFIG.persist:
            print_header(f"Deleting instance {instance_id} (from local)...")
            utils.delete_instance(instance_id)

        sys.exit(1)


def run_commands_on_instance(commands, instance_ids):
    """Runs commands on remote linux instances
    :param commands: a list of strings, each one a command to execute on the instances
    :param instance_ids: a list of instance_id strings, of the instances on which
                         to execute the command
    :return: the response from the send_command function (check the boto3 docs
             for ssm client.send_command() )
    """

    ssm = CONFIG.get_session().client("ssm")
    resp = ssm.send_command(
        DocumentName="AWS-RunShellScript",  # One of AWS' preconfigured documents
        Parameters={"commands": commands},
        InstanceIds=instance_ids,
    )
    return resp



import logging
import subprocess
from os.path import join

from botocore.exceptions import ClientError

from nimbo import CONFIG
from nimbo.core.print import print


def s3_cp_command(source, target, delete=False):
    command = (
        f"aws s3 cp {source} {target} "
        f" --profile {CONFIG.aws_profile} --region {CONFIG.region_name}"
    )

    if delete:
        command += " --delete"

    if CONFIG.encryption:
        command += f" --sse {CONFIG.encryption}"
    return command


def s3_sync_command(source, target, delete=False):
    command = (
        f"aws s3 sync {source} {target} "
        f" --profile {CONFIG.aws_profile} --region {CONFIG.region_name}"
    )

    if delete:
        command += " --delete"

    if CONFIG.encryption:
        command += f" --sse {CONFIG.encryption}"
    return command


def upload_file(file_name, bucket, object_name=None):
    """Upload a file to an S3 bucket

    :param file_name: File to upload
    :param bucket: Bucket to upload to
    :param object_name: S3 object name. If not specified then file_name is used
    :return: True if file was uploaded, else False
    """

    # If S3 object_name was not specified, use file_name
    if object_name is None:
        object_name = file_name

    # Upload the file
    s3 = CONFIG.get_session().client("s3")
    try:
        s3.upload_file(file_name, bucket, object_name)
    except ClientError as e:
        print(e, style="error")
        return False
    return True


def create_bucket(bucket_name, dry_run=False):
    """Create an S3 bucket in a specified region

    :param bucket_name: Bucket to create
    :param dry_run
    :return: True if bucket created, else False
    """

    try:
        session = CONFIG.get_session()
        s3 = session.client("s3")
        location = {"LocationConstraint": session.region_name}
        s3.create_bucket(Bucket=bucket_name, CreateBucketConfiguration=location)
    except ClientError as e:
        if e.response["Error"]["Code"] == "BucketAlreadyOwnedByYou":
            print("Bucket nimbo-main-bucket already exists.", style="warning")
        else:
            print(e, style="error")
        return False

    print("Bucket %s created." % bucket_name)
    return True


def list_buckets():
    s3 = CONFIG.get_session().client("s3")
    response = s3.list_buckets()

    print("Existing buckets:")
    for bucket in response["Buckets"]:
        print(f' {bucket["Name"]}')


def list_snapshots():
    # Retrieve the list of existing buckets
    ec2 = CONFIG.get_session().client("ec2")

    response = ec2.describe_snapshots(
        Filters=[{"Name": "tag:created_by", "Values": ["nimbo"]}],
        MaxResults=100,
    )
    return list(sorted(response["Snapshots"], key=lambda x: x["StartTime"]))


def check_snapshot_state(snapshot_id):
    ec2 = CONFIG.get_session().client("ec2")
    response = ec2.describe_snapshots(SnapshotIds=[snapshot_id])
    return response["Snapshots"][0]["State"]


def sync_folder(source, target, delete=False):
    command = s3_sync_command(source, target, delete)
    print(f"Running command: {command}")
    subprocess.Popen(command, shell=True).communicate()


# noinspection DuplicatedCode
def pull(folder, delete=False):
    assert folder in ["datasets", "results", "logs"]

    if folder == "logs":
        source = join(CONFIG.s3_results_path, "nimbo-logs")
        target = join(CONFIG.local_results_path, "nimbo-logs")
    else:
        if folder == "results":
            source = CONFIG.s3_results_path
            target = CONFIG.local_results_path
        else:
            source = CONFIG.s3_datasets_path
            target = CONFIG.local_datasets_path

    sync_folder(source, target, delete)


# noinspection DuplicatedCode
def push(folder, delete=False):
    assert folder in ["datasets", "results", "logs"]

    if folder == "logs":
        source = join(CONFIG.local_results_path, "nimbo-logs")
        target = join(CONFIG.s3_results_path, "nimbo-logs")
    else:
        if folder == "results":
            source = CONFIG.local_results_path
            target = CONFIG.s3_results_path
        else:
            source = CONFIG.local_datasets_path
            target = CONFIG.s3_datasets_path

    sync_folder(source, target, delete)


def ls(path):
    profile = CONFIG.aws_profile
    region = CONFIG.region_name
    path = path.rstrip("/") + "/"
    command = f"aws s3 ls {path} --profile {profile} --region {region}"
    print(f"Running command: {command}")
    subprocess.Popen(command, shell=True).communicate()
